import asyncio
from inspect import isawaitable
from traceback import format_exc
from types import FunctionType

from .config import Config
from .exceptions import Handler
from .log import log, logging
from .middleware import Middleware
from .response import HTTPResponse
from .router import Router
from .server import serve
from .exceptions import ServerError


class Sanic:
    def __init__(self, name, router=None, error_handler=None):
        self.name = name
        self.router = router or Router()
        self.router = router or Router()
        self.error_handler = error_handler or Handler(self)
        self.config = Config()
        self.request_middleware = []
        self.response_middleware = []

    # -------------------------------------------------------------------- #
    # Registration
    # -------------------------------------------------------------------- #

    # Decorator
    def route(self, uri, methods=None):
        """
        Decorates a function to be registered as a route
        :param uri: path of the URL
        :param methods: list or tuple of methods allowed
        :return: decorated function
        """

        def response(handler):
            self.router.add(uri=uri, methods=methods, handler=handler)
            return handler

        return response

    # Decorator
    def exception(self, *exceptions):
        """
        Decorates a function to be registered as a route
        :param uri: path of the URL
        :param methods: list or tuple of methods allowed
        :return: decorated function
        """

        def response(handler):
            for exception in exceptions:
                self.error_handler.add(exception, handler)
            return handler

        return response

    # Decorator
    def middleware(self, *args, **kwargs):
        """
        Decorates and registers middleware to be called before a request
        can either be called as @app.middleware or @app.middleware('request')
        """
        middleware = None
        attach_to = 'request'

        def register_middleware(middleware):
            if attach_to == 'request':
                self.request_middleware.append(middleware)
            if attach_to == 'response':
                self.response_middleware.append(middleware)
            return middleware

        # Detect which way this was called, @middleware or @middleware('AT')
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            return register_middleware(args[0])
        else:
            attach_to = args[0]
            return register_middleware

        if isinstance(middleware, FunctionType):
            middleware = Middleware(process_request=middleware)

        return middleware

    # -------------------------------------------------------------------- #
    # Request Handling
    # -------------------------------------------------------------------- #

    async def handle_request(self, request, response_callback):
        """
        Takes a request from the HTTP Server and returns a response object to be sent back
        The HTTP Server only expects a response object, so exception handling must be done here
        :param request: HTTP Request object
        :param response_callback: Response function to be called with the response as the only argument
        :return: Nothing
        """
        try:
            # Middleware process_request
            response = False
            # The if improves speed.  I don't know why
            if self.request_middleware:
                for middleware in self.request_middleware:
                    response = middleware(request)
                    if isawaitable(response):
                        response = await response
                    if response:
                        break

            # No middleware results
            if not response:
                # Fetch handler from router
                handler, args, kwargs = self.router.get(request)
                if handler is None:
                    raise ServerError("'None' was returned while requesting a handler from the router")

                # Run response handler
                response = handler(request, *args, **kwargs)
                if isawaitable(response):
                    response = await response

                # Middleware process_response
                if self.response_middleware:
                    for middleware in self.response_middleware:
                        _response = middleware(request, response)
                        if isawaitable(_response):
                            _response = await _response
                        if _response:
                            response = _response
                            break

        except Exception as e:
            try:
                response = self.error_handler.response(request, e)
                if isawaitable(response):
                    response = await response
            except Exception as e:
                if self.debug:
                    response = HTTPResponse("Error while handling error: {}\nStack: {}".format(e, format_exc()))
                else:
                    response = HTTPResponse("An error occured while handling an error")

        response_callback(response)

    # -------------------------------------------------------------------- #
    # Execution
    # -------------------------------------------------------------------- #

    def run(self, host="127.0.0.1", port=8000, debug=False, after_start=None, before_stop=None):
        """
        Runs the HTTP Server and listens until keyboard interrupt or term signal.
        On termination, drains connections before closing.
        :param host: Address to host on
        :param port: Port to host on
        :param debug: Enables debug output (slows server)
        :param after_start: Function to be executed after the server starts listening
        :param before_stop: Function to be executed when a stop signal is received before it is respected
        :return: Nothing
        """
        self.error_handler.debug = True
        self.debug = debug

        if debug:
            log.setLevel(logging.DEBUG)
        log.debug(self.config.LOGO)

        # Serve
        log.info('Goin\' Fast @ http://{}:{}'.format(host, port))

        try:
            serve(
                host=host,
                port=port,
                debug=debug,
                after_start=after_start,
                before_stop=before_stop,
                request_handler=self.handle_request,
                request_timeout=self.config.REQUEST_TIMEOUT,
                request_max_size=self.config.REQUEST_MAX_SIZE,
            )
        except:
            pass

    def stop(self):
        """
        This kills the Sanic
        """
        asyncio.get_event_loop().stop()



import ujson
import httptools
from ujson import loads as json_loads
from urllib.parse import parse_qs

STATUS_CODES = {
    200: 'OK',
    400: 'Bad Request',
    401: 'Unauthorized',
    402: 'Payment Required',
    403: 'Forbidden',
    404: 'Not Found',
    405: 'Method Not Allowed',
    500: 'Internal Server Error',
    501: 'Not Implemented',
    502: 'Bad Gateway',
    503: 'Service Unavailable',
    504: 'Gateway Timeout',
}


class HTTPResponse:
    __slots__ = ('body', 'status', 'content_type', 'headers')

    def __init__(self, body=None, status=200, headers=[], content_type='text/plain', body_bytes=b''):
        self.content_type = content_type

        if not body is None:
            self.body = body.encode('utf-8')
        else:
            self.body = body_bytes

        self.status = status
        self.headers = headers

    def output(self, version="1.1", keep_alive=False, keep_alive_timeout=None):
        # This is all returned in a kind-of funky way
        # We tried to make this as fast as possible in pure python
        additional_headers = []
        if keep_alive and not keep_alive_timeout is None:
            additional_headers = [b'Keep-Alive: timeout=', str(keep_alive_timeout).encode(), b's\r\n']
        if self.headers:
            for name, value in self.headers.items():
                additional_headers.append('{}: {}\r\n'.format(name, value).encode('utf-8'))

        return b''.join([
                            'HTTP/{} {} {}\r\n'.format(version, self.status,
                                                       STATUS_CODES.get(self.status, 'FAIL')).encode(),
                            b'Content-Type: ', self.content_type.encode(), b'\r\n',
                            b'Content-Length: ', str(len(self.body)).encode(), b'\r\n',
                            b'Connection: ', ('keep-alive' if keep_alive else 'close').encode(), b'\r\n',
                        ] + additional_headers + [
                            b'\r\n',
                            self.body,
                        ])


def json(body, status=200, headers=None):
    return HTTPResponse(ujson.dumps(body), headers=headers, status=status,
                        content_type="application/json; charset=utf-8")


def text(body, status=200, headers=None):
    return HTTPResponse(body, status=status, headers=headers, content_type="text/plain; charset=utf-8")


def html(body, status=200, headers=None):
    return HTTPResponse(body, status=status, headers=headers, content_type="text/html; charset=utf-8")



import re
from collections import namedtuple
from .exceptions import NotFound, InvalidUsage

Route = namedtuple("Route", ['handler', 'methods', 'pattern', 'parameters'])
Parameter = namedtuple("Parameter", ['name', 'cast'])


class Router:
    """
    Router supports basic routing with parameters and method checks
    Usage:
        @sanic.route('/my/url/<my_parameter>', methods=['GET', 'POST', ...])
        def my_route(request, my_parameter):
            do stuff...

    Parameters will be passed as keyword arguments to the request handling function provided
    Parameters can also have a type by appending :type to the <parameter>.  If no type is provided,
        a string is expected.  A regular expression can also be passed in as the type

    TODO:
        This probably needs optimization for larger sets of routes,
        since it checks every route until it finds a match which is bad and I should feel bad
    """
    routes = None
    regex_types = {
        "string": (None, "\w+"),
        "int": (int, "\d+"),
        "number": (float, "[0-9\\.]+"),
        "alpha": (None, "[A-Za-z]+"),
    }

    def __init__(self):
        self.routes = []

    def add(self, uri, methods, handler):
        """
        Adds a handler to the route list
        :param uri: Path to match
        :param methods: Array of accepted method names.  If none are provided, any method is allowed
        :param handler: Request handler function.  When executed, it should provide a response object.
        :return: Nothing
        """

        # Dict for faster lookups of if method allowed
        methods_dict = {method: True for method in methods} if methods else None

        parameters = []

        def add_parameter(match):
            # We could receive NAME or NAME:PATTERN
            parts = match.group(1).split(':')
            if len(parts) == 2:
                parameter_name, parameter_pattern = parts
            else:
                parameter_name = parts[0]
                parameter_pattern = 'string'

            # Pull from pre-configured types
            parameter_regex = self.regex_types.get(parameter_pattern)
            if parameter_regex:
                parameter_type, parameter_pattern = parameter_regex
            else:
                parameter_type = None

            parameter = Parameter(name=parameter_name, cast=parameter_type)
            parameters.append(parameter)

            return "({})".format(parameter_pattern)

        pattern_string = re.sub("<(.+?)>", add_parameter, uri)
        pattern = re.compile("^{}$".format(pattern_string))

        route = Route(handler=handler, methods=methods_dict, pattern=pattern, parameters=parameters)
        self.routes.append(route)

    def get(self, request):
        """
        Gets a request handler based on the URL of the request, or raises an error
        :param request: Request object
        :return: handler, arguments, keyword arguments
        """

        route = None
        args = []
        kwargs = {}
        for _route in self.routes:
            match = _route.pattern.match(request.url)
            if match:
                for index, parameter in enumerate(_route.parameters, start=1):
                    value = match.group(index)
                    kwargs[parameter.name] = parameter.cast(value) if parameter.cast is not None else value
                route = _route
                break

        if route:
            if route.methods and not request.method in route.methods:
                raise InvalidUsage("Method {} not allowed for URL {}".format(request.method, request.url),
                                   status_code=405)
            return route.handler, args, kwargs
        else:
            raise NotFound("Requested URL {} not found".format(request.url))


class SimpleRouter:
    """
    Simple router records and reads all routes from a dictionary
    It does not support parameters in routes, but is very fast
    """
    routes = None

    def __init__(self):
        self.routes = {}

    def add(self, uri, methods, handler):
        # Dict for faster lookups of method allowed
        methods_dict = {method: True for method in methods} if methods else None
        self.routes[uri] = Route(handler=handler, methods=methods_dict, pattern=uri, parameters=None)

    def get(self, request):
        route = self.routes.get(request.url)
        if route:
            if route.methods and not request.method in route.methods:
                raise InvalidUsage("Method {} not allowed for URL {}".format(request.method, request.url),
                                   status_code=405)
            return route.handler, [], {}
        else:
            raise NotFound("Requested URL {} not found".format(request.url))
